from apix.explore import AsyncExplorer


class APILister:
    def __init__(self, settings):
        self.explorer = AsyncExplorer(
            name=settings.candore.product_name,
            version=settings.candore.product_version,
            host_url=settings.candore.base_url,
            base_path=settings.candore.docpath,
            parser=settings.candore.parser,
        )
        self.list_endpoints = {}

    def _endpoints(self):
        self.explorer.explore()
        self.yaml_data = self.explorer.yaml_data()
        return self.yaml_data

    def lister_endpoints(self):
        list_endpoints = {}

        for component, data in self._endpoints().items():
            methods = data.get("methods", [])
            component_list_apis = [
                path.lstrip("GET ")
                for method in methods
                for path in method.get("index", {}).get("paths", [])
                if "id" not in path
            ]
            list_endpoints[component] = component_list_apis

        self.list_endpoints = list_endpoints
        return list_endpoints



from flask import Flask
from flask import render_template

app = Flask(__name__)


# Define a route to display the JSON data in a table
@app.route("/")
def display_json_table(results_json=None):
    # Fetch JSON data from your module (replace with your data source)

    return render_template("table.html", data=results_json)


def render_webpage():
    app.run(debug=True)



"""Explore and API and save the results."""
import asyncio
import time
from pathlib import Path

import aiohttp
import attr
import requests
import yaml
from logzero import logger

from apix.parsers import apipie
from apix.parsers import test


@attr.s()
class AsyncExplorer:
    name = attr.ib(default=None)
    version = attr.ib(default=None)
    host_url = attr.ib(default=None)
    base_path = attr.ib(default=None)
    parser = attr.ib(default=None)
    data_dir = attr.ib(default=None)
    compact = attr.ib(default=False)
    _data = attr.ib(default=attr.Factory(dict), repr=False)
    _queue = attr.ib(default=attr.Factory(list), repr=False)

    def __attrs_post_init__(self):
        """perform the more complex steps of class initialization"""
        if not self.version:
            self.version = time.strftime("%Y-%m-%d", time.localtime())
        # choose the correct parser class from known parsers
        if self.parser.lower() == "apipie":
            self.parser = apipie.APIPie()
        elif self.parser.lower() == "test":
            self.parser = test.TestParser()
        if not self.parser or isinstance(self.parser, str):
            logger.warning("No known parser specified! Please review documentation.")

    async def _async_get(self, session, link):
        """
        visit a page and download the content returning the link and content
        """
        async with session.get(self.host_url + link[1], ssl=False) as response:
            content = await response.read()
            logger.debug(link[1])
            return (link, content)

    async def _async_loop(self, links):
        """asynchronously visit each stored link and store results"""
        tasks = []
        async with aiohttp.ClientSession() as session:
            for link in links:
                task = asyncio.ensure_future(self._async_get(session, link))
                tasks.append(task)
            results = await asyncio.gather(*tasks)
            for result in results:
                self._queue.append(result)

    def _visit_links(self, links, retries=3):
        """
        Main controller for asynchronous page visiting, will attempt 3 retries
        """
        try:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self._async_loop(links))
        except aiohttp.client_exceptions.ServerDisconnectedError:
            logger.warning(
                "Lost connection to host.".join("Retrying in 10 seconds" if retries else "")
            )
            if retries:
                time.sleep(10)
                self._visit_links(links, retries - 1)

    def _link_params(self):
        """scrape the content downloaded from each page"""
        while self._queue:
            link, content = self._queue.pop(0)
            logger.debug(f"Scraping {link[1]}")
            self._data[link[1]] = self.parser.scrape_content(content)

    def yaml_data(self):
        yaml_data = self.parser.yaml_format(self._data)
        if not yaml_data:
            logger.warning("No data to be saved. Exiting.")
            return
        return yaml_data

    def save_data(self, return_path=False):
        """convert the stored data into yaml-friendly dict and save"""
        yaml_data = self.yaml_data()
        if self.compact:
            from apix.diff import VersionDiff

            yaml_data = VersionDiff._truncate(yaml_data)
            fpath = Path(f"{self.data_dir}APIs/{self.name}/{self.version}-comp.yaml")
        else:
            fpath = Path(f"{self.data_dir}APIs/{self.name}/{self.version}.yaml")
        if fpath.exists():
            logger.warning(f"{fpath} already exists. Deleting..")
            fpath.unlink()
        # create the directory, if it doesn't exist
        fpath.parent.mkdir(parents=True, exist_ok=True)
        fpath.touch()
        logger.info(f"Saving results to {fpath}")
        with fpath.open("w+") as outfile:
            yaml.dump(yaml_data, outfile, default_flow_style=False)
        if return_path:
            return fpath

    def explore(self):
        """main function for the explore module
        visit the initial page, pulling all links from the page
        grab content at link found
        sort the links to account for incorrect async return order
        scrape the content at each page, per the parser's direction
        """
        result = requests.get(self.host_url + self.base_path, verify=False)
        if not result:
            logger.warning(
                f"I couldn't find anything useful at " f"{self.host_url}{self.base_path}."
            )
            return
        self.base_path = self.base_path.replace(".html", "")  # for next step
        logger.info(f"Starting to explore {self.host_url}{self.base_path}")
        if hasattr(self.parser, "pull_links"):
            links = self.parser.pull_links(result, self.base_path)
            logger.debug(f"Found {len(links)} links!")
            self._visit_links(links)
            # sort the results by link name, to normalize return order
            self._queue = sorted(self._queue, key=lambda x: x[0][1])
            self._link_params()
        else:
            self.parser.scrape_content(result)
        return True



"""
A module responsible for calculating expected and skipped variations from
`conf/variations` yaml file and convert them into processable list
"""
from functools import cached_property
from pathlib import Path

import yaml


class Variations:
    def __init__(self, settings):
        self.settings = settings

    def get_paths(self, variations, prefix="", separator="/"):
        paths = []
        if isinstance(variations, dict):
            for key, value in variations.items():
                paths.extend(self.get_paths(value, f"{prefix}{key}{separator}"))
        elif isinstance(variations, list):
            for item in variations:
                paths.extend(self.get_paths(item, prefix, separator))
        else:
            paths.append(f"{prefix}{variations}")

        return paths

    @cached_property
    def variations(self):
        templates_path = Path(self.settings.candore.var_file)
        if not templates_path.exists():
            print(f"The file {templates_path} does not exist.")
        with templates_path.open() as yaml_file:
            yaml_data = yaml.safe_load(yaml_file)
        return yaml_data

    @cached_property
    def expected_variations(self):
        return self.get_paths(variations=self.variations.get("expected_variations"))

    @cached_property
    def skipped_variations(self):
        return self.get_paths(variations=self.variations.get("skipped_variations"))



import json


class Finder:
    def read_json(self, jsonfile=None):
        jdata = None
        with open(jsonfile) as file:
            jdata = json.load(file)
        return jdata

    def json_iterator(self, data=None, path=None, delimiter=None):
        if not data:
            return data
        if path:
            path_contents = path.strip().split(delimiter)
            first_path = path_contents[0]
            remaining_path = f'{delimiter}'.join(path_contents[1:])

            if isinstance(data, dict):
                if first_path in data:
                    return self.json_iterator(data.get(first_path, None), remaining_path, delimiter)
                else:
                    print(
                        f'Oh!O! {first_path} is not found in given path, please correct the path!'
                    )
            elif isinstance(data, list):
                for index, element_data in enumerate(data):
                    if isinstance(element_data, dict):
                        if 'id' in element_data and str(element_data['id']) == first_path:
                            return self.json_iterator(element_data, remaining_path, delimiter)
                    elif isinstance(element_data, list):
                        if first_path in element_data:
                            return self.json_iterator(
                                element_data[index], remaining_path, delimiter
                            )
        else:
            return data

    def find(self, path=None, json_file=None, delimiter='/'):
        jdata = self.read_json(json_file)
        return self.json_iterator(jdata, path=path, delimiter=delimiter)



# -*- encoding: utf-8 -*-
"""A collection of miscellaneous helpers that don't quite fit in."""
from copy import deepcopy
from pathlib import Path

import yaml
from logzero import logger


def get_api_list(data_dir=None, mock=False):
    """Return a list of saved apis, if they exist"""
    api_dir = Path(f"{data_dir}APIs/" if not mock else f"{data_dir}tests/APIs/")
    # check exists
    if not api_dir.exists():
        return None
    # get all versions in directory, that aren't diffs
    apis = [(api.name, api.stat().st_mtime) for api in api_dir.iterdir() if api.is_dir()] or []
    apis = [api for api, _ in sorted(apis, key=lambda x: x[1], reverse=True)]
    return apis


def get_ver_list(api_name, data_dir=None, mock=False):
    """Return a list of saved api versions, if they exist"""
    if mock:
        save_path = Path(f"{data_dir}tests/APIs/{api_name}")
    else:
        save_path = Path(f"{data_dir}APIs/{api_name}")
    # check exists
    if not save_path.exists():
        return None
    # get all versions in directory, that aren't diffs
    versions = [
        v_file.name.replace(".yaml", "")
        for v_file in save_path.iterdir()
        if "-diff." not in v_file.name and "-comp." not in v_file.name and ".yaml" in v_file.name
    ] or []
    return sorted(versions, reverse=True)


def get_latest(api_name=None, data_dir=None, mock=False):
    """Get the latest api version, if it exists"""
    if not api_name:
        return get_api_list(data_dir, mock=mock)[0]
    else:
        ver_list = get_ver_list(api_name, data_dir, mock=mock) or [None]
        return ver_list[0]


def get_previous(api_name, version, data_dir=None, mock=False):
    """Get the api version before `version`, if it isn't last"""
    api_list = get_ver_list(api_name, data_dir, mock=mock)
    if api_list and version in api_list:
        v_pos = api_list.index(version)
        if v_pos + 2 <= len(api_list):
            return api_list[v_pos + 1]
    return None


def shift_text(text, shift=1):
    """Shifts blocks or a single line of text by 4 * shift spaces"""
    new_text = ""
    if "\n" in text:
        for line in text.split("\n"):
            new_text += "    " * shift + line + "\n"
    else:
        new_text = "    " * shift + text
    return new_text


def load_api(api_name, version, data_dir=None, mock=False):
    """Load the saved yaml to dict, if the file exists"""
    if mock:
        a_path = Path(f"{data_dir}tests/APIs/{api_name}/{version}.yaml")
    else:
        a_path = Path(f"{data_dir}APIs/{api_name}/{version}.yaml")
    if not a_path.exists():
        return None
    logger.info(f"Loading {api_name} v{version} from {a_path}")
    return yaml.load(a_path.open("r"), Loader=yaml.Loader) or None


def save_api(api_name, version, api_dict, data_dir=None, compact=False, mock=False):
    """Save the dict to yaml, if the file doesn't exist"""
    if mock:
        a_path = Path(
            f"{data_dir}tests/APIs/{api_name}/{version}" f"{'-comp' if compact else ''}.yaml"
        )
    else:
        a_path = Path(f"{data_dir}APIs/{api_name}/{version}" f"{'-comp' if compact else ''}.yaml")
    a_path.parent.mkdir(parents=True, exist_ok=True)
    logger.info(f"Saving {api_name} v{version} to {a_path}")
    with a_path.open("w") as f:
        yaml.dump(api_dict, f, default_flow_style=False)


def merge_dicts(dict1, dict2):
    """Merge two nested dicitonaries together"""
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return dict1
    merged = {}
    dupe_keys = dict1.keys() & dict2.keys()
    for key in dupe_keys:
        merged[key] = merge_dicts(dict1[key], dict2[key])
    for key in dict1.keys() - dupe_keys:
        merged[key] = deepcopy(dict1[key])
    for key in dict2.keys() - dupe_keys:
        merged[key] = deepcopy(dict2[key])
    return merged



from pathlib import Path
from pathlib import PurePath

from dynaconf import Dynaconf
from dynaconf.validator import Validator

CURRENT_DIRECTORY = Path().resolve()


def candore_settings(option_settings_file=None, option_components_file=None):
    settings_file = (
        PurePath(option_settings_file)
        if option_settings_file
        else PurePath(CURRENT_DIRECTORY, "settings.yaml")
    )
    components_file = (
        PurePath(option_components_file)
        if option_components_file
        else PurePath(CURRENT_DIRECTORY, "components.yaml")
    )
    # Initialize and Configure Settings
    settings = Dynaconf(
        core_loaders=["YAML"],
        envvar_prefix="CANDORE",
        settings_files=[settings_file, components_file],
        envless_mode=True,
        lowercase_read=True,
    )
    validate_settings(settings)
    return settings


def validate_settings(settings):
    provider_settings = [
        f"candore.{setting_key}" for setting_key in settings.to_dict().get("CANDORE")
    ]
    settings.validators.register(Validator(*provider_settings, ne=None))
    try:
        settings.validators.validate()
    except Exception as ecc:
        raise ecc



import asyncio  # noqa: F401
from functools import cached_property

import aiohttp


class Extractor:
    def __init__(self, settings, apilister=None):
        """Extract and save data using API lister endpoints

        :param apilister: APILister object
        """
        self.settings = settings
        self.username = self.settings.candore.username
        self._passwd = self.settings.candore.password
        self.base = self.settings.candore.base_url
        self.verify_ssl = False
        self.auth = aiohttp.BasicAuth(self.username, self._passwd)
        self.connector = aiohttp.TCPConnector(ssl=self.verify_ssl)
        self.client = None
        self.apilister = apilister
        self.full = False

    @cached_property
    def dependent_components(self):
        if hasattr(self.settings, "components"):
            return self.settings.components.dependencies

    @cached_property
    def ignore_components(self):
        if hasattr(self.settings, "components"):
            return self.settings.components.ignore

    @cached_property
    def api_endpoints(self):
        return self.apilister.lister_endpoints()

    async def _start_session(self):
        if not self.client:
            self.client = aiohttp.ClientSession(auth=self.auth, connector=self.connector)
        return self.client

    async def _end_session(self):
        await self.client.close()

    async def __aenter__(self):
        await self._start_session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._end_session()

    async def paged_results(self, **get_params):
        async with self.client.get(**get_params) as response:
            if response.status == 200:
                _paged_results = await response.json()
                _paged_results = _paged_results.get("results")
                return _paged_results

    async def fetch_component_entities(self, **comp_params):
        entity_data = []
        endpoint = comp_params.get("endpoint", None)
        data = comp_params.get("data")
        dependency = comp_params.get("dependency", None)
        _request = {"url": self.base + "/" + endpoint, "params": {}}
        if data and dependency:
            _request["params"].update({f"{dependency}_id": data})
        async with self.client.get(**_request) as response:
            if response.status == 200:
                results = await response.json()
                if "results" in results:
                    entities = results.get("results")
                    entity_data.extend(entities)
                else:
                    # Return an empty directory for endpoints
                    # like services, api etc
                    # which does not have results
                    return entity_data
            else:
                return entity_data
        # If the entity has multiple pages, fetch them all
        if self.full:
            total_pages = results.get("total") // results.get("per_page") + 1
            if total_pages > 1:
                print(
                    f"Endpoint {endpoint} has {total_pages} pages. "
                    "This would take some time ...."
                )
                for page in range(2, total_pages + 1):
                    _request["params"].update({"page": page})
                    page_entities = await self.paged_results(**_request)
                    entity_data.extend(page_entities)
        return entity_data

    async def dependency_ids(self, dependency):
        # All the Ids of a specific dependency
        # e.g Organization IDs 1, 2, 3, 4
        endpoint = self.api_endpoints[f"{dependency}s"][0]
        depe_lists = await self.fetch_component_entities(endpoint=endpoint)
        depen_ids = [dep_dict["id"] for dep_dict in depe_lists]
        return depen_ids

    async def component_params(self, component_endpoint):
        """
        component_endpoints = ['katello/api/activationkeys']
        endpoints = ['activationkeys']
        :param component_endpoints:
        :return:
        """
        data = {}
        dependency = None
        # remove ignored endpoints
        _last = component_endpoint.rsplit("/")[-1]
        # Ignorable endpoint
        if self.ignore_components and _last in self.ignore_components:
            return
        # Return results for components those has dependencies
        if self.dependent_components and _last in self.dependent_components.keys():
            dependency = self.dependent_components[_last]
            data = await self.dependency_ids(dependency)
        return {"endpoint": component_endpoint, "data": data, "dependency": dependency}

    async def process_entities(self, endpoints):
        """
        endpoints = ['katello/api/actiovationkeys']
        """
        comp_data = []
        entities = None
        for endpoint in endpoints:
            comp_params = await self.component_params(component_endpoint=endpoint)
            if comp_params:
                entities = []
                if isinstance(comp_params.get("data"), list):
                    for data_point in comp_params.get("data"):
                        depen_data = await self.fetch_component_entities(
                            endpoint=comp_params["endpoint"],
                            dependency=comp_params.get("dependency"),
                            data=data_point,
                        )
                        if not depen_data:
                            continue
                        entities.extend(depen_data)
                else:
                    entities = await self.fetch_component_entities(**comp_params)
            if entities:
                comp_data.extend(entities)
        return comp_data

    async def extract_all_entities(self):
        """Extract all entities fom all endpoints

        :return:
        """
        all_data = {}
        for component, endpoints in self.api_endpoints.items():
            if endpoints:
                comp_entities = await self.process_entities(endpoints=endpoints)
                all_data[component] = comp_entities
        return all_data
